print("import: 'textacy'")
import textacy

print("import: 'textacy.augmentation'")
import textacy.augmentation

print("import: 'textacy.cache'")
import textacy.cache

print("import: 'textacy.constants'")
import textacy.constants

print("import: 'textacy.corpus'")
import textacy.corpus

print("import: 'textacy.datasets'")
import textacy.datasets

print("import: 'textacy.errors'")
import textacy.errors

print("import: 'textacy.extensions'")
import textacy.extensions

print("import: 'textacy.extract'")
import textacy.extract

print("import: 'textacy.io'")
import textacy.io

print("import: 'textacy.lang_id'")
import textacy.lang_id

print("import: 'textacy.preprocessing'")
import textacy.preprocessing

print("import: 'textacy.representations'")
import textacy.representations

print("import: 'textacy.resources'")
import textacy.resources

print("import: 'textacy.similarity'")
import textacy.similarity

print("import: 'textacy.spacier'")
import textacy.spacier

print("import: 'textacy.text_stats'")
import textacy.text_stats

print("import: 'textacy.tm'")
import textacy.tm

print("import: 'textacy.types'")
import textacy.types

print("import: 'textacy.utils'")
import textacy.utils

print("import: 'textacy.viz'")
import textacy.viz

